/*
 * Ground up (well, except for name unmangling) re-implementation of dumpbin.
 *
 * This is really more of an investigative project than a serious attempt at a tool, however it exports
 * the resultant data structures so that a console or gui app can use them.
 *
 * In order for me to validate against dumpbin (though that's not without bugs of its own ;) - the header
 * dumped to the console and the error messages are identical.
 *
 * (originally this just dumped everything to stdout, and I've not removed the dumping - the wumpbin gui
 * which sits on top of this just doesn't have much of a stdout attached ;)
 *
 * lee@benf.org
 */
#include "stdafx.h"
#include "dumpbin.h"
#include "hexdump.h"
#include "exceptions.h"
#include "unmangle.h"
#include "mmf.h"
#include "processed.h"
#include "processed_impl.h"
#include "dump_helpers.h"
#include <winnt.h>
#include <string>
#include <sstream>
#include <map>
#include <vector>
#include <set>

class result_heirachy_error : public result_heirachy
{
private:
	const std::string m_etext;
public:
	result_heirachy_error(std::string err) : m_etext(err) {};
	virtual bool is_error() const { return true; }
	virtual std::string error_text() { return m_etext; }
	virtual class dbobj * get_node(int node_id) { return NULL; }
	virtual std::vector<int> get_children(int node_id) { return std::vector<int>(); }
};


class result_heirachy_impl : public result_heirachy
{
	memory_mapped_file::ptr m_filedata;
		
	int m_curr_leaf;
	std::vector<dbobj::ptr> m_data;
	std::map<int, std::vector<int> > m_relations;
	std::map<std::string, int> m_redirects;
	friend class rhi_stack;
	
public:
	result_heirachy_impl() : m_curr_leaf(-1)
	{
	}
	virtual bool is_error() const 
	{
		return false;
	}
	virtual std::string error_text() { return ""; }
	virtual class dbobj * get_node(int node_id) 
	{
		dbobj * ret = m_data[node_id].get();
		std::string indirkey;
		while (ret->is_indirection(indirkey)) 
		{
			std::map<std::string, int>::iterator mit = m_redirects.find(indirkey);
			if (mit == m_redirects.end())
			{
				return ret;
			}
			else
			{
				ret = m_data[mit->second].get();
			}
		}
		return ret;
	}
	virtual std::vector<int> get_children(int node_id) 
	{
		// fixme - mapref waste.
		return m_relations[node_id];
	}

public:
	int add_node(dbobj::ptr & p)
	{
		int x = (int)m_data.size();
		m_data.push_back(p);
		if (m_curr_leaf >= 0) m_relations[m_curr_leaf].push_back(x);
		return x;
	}
	int add_node(std::string redirect_key, dbobj::ptr & p)
	{
		int x = (int)m_data.size();
		m_data.push_back(p);
		if (m_curr_leaf >= 0) m_relations[m_curr_leaf].push_back(x);
		m_redirects[redirect_key] = x;
		return x;
	}
	memory_mapped_file::ptr get_filedata() { return m_filedata; }
	void set_filedata(memory_mapped_file::ptr & f) { m_filedata = f; }
protected:
	int set_proc_leaf(int newleaf)
	{
		int x = m_curr_leaf;
		m_curr_leaf = newleaf;
		return x;
	}
};

class rhi_stack
{
	result_heirachy_impl * m_prhi;
	int m_old_leaf;
	rhi_stack(const rhi_stack & other);
	rhi_stack & operator=(const rhi_stack & other);
public:
	rhi_stack(result_heirachy_impl * r, int thisl)
	{
		m_prhi = r;
		m_old_leaf = r->set_proc_leaf(thisl);
	}
	~rhi_stack() 
	{
		m_prhi->set_proc_leaf(m_old_leaf);
	}
};

std::string image_header_file_type(IMAGE_FILE_HEADER * h)
{
	unsigned long chars = h->Characteristics;
	if (chars & IMAGE_FILE_DLL) return "File Type: DLL\n\n";
	if (chars & IMAGE_FILE_EXECUTABLE_IMAGE) return "File Type: EXECUTABLE IMAGE\n\n";
	return "";
}

void dump_coff_header(IMAGE_FILE_HEADER & h, result_heirachy_impl * prh)
{
	dbobj_coff_header * dco = new dbobj_coff_header(&h);
	dbobj::ptr dbp(dco);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dco->get_display();
}



void cook_section_information(IMAGE_SECTION_HEADER & h, int id, cofpcdat & cd)
{
	cd.add_section_information(&h, id);
}

/*
 * We have to refer forward to the symbol table sometimes -  
 *
 *         6    5 00012B3B __xmlBufferAllocScheme
 *         7    6 00012B95 __xmlDefaultBufferSize = _xmlHashQLookup
 *         8    7 00012BEF __xmlDefaultSAXHandler
 *----->
 *  0F2 00012B95 SECT1  notype ()    External     | _xmlHashQLookup
 *
 */
void dump_export_data_directory(byte * pd, int size, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_export_directory * dso = new dbobj_export_directory(pd, size, &cd);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD0", dbp));

	std::cout << dso->get_display();
}


void dump_import_data_directory(byte * pd, int size, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_import_directory * dso = new dbobj_import_directory(pd, size, &cd);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD1", dbp));

	std::cout << dso->get_display();
}

/*
 * See winnt - the exception table is repurposed for wince to be a function table.
 * This is the handler for wince - we should only really use it when appropriate...... (FIXME).
 */
void dump_exception_data_directory(byte * start, int size, image_section_helper & sh, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_exception_directory * dso = new dbobj_exception_directory(start, size, &cd);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD3", dbp));

	std::cout << dso->get_display();
}

void dump_load_config_data_directory(byte * pd, int size, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_load_config_directory * dso = new dbobj_load_config_directory(pd, size, &cd);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD10", dbp));

	std::cout << dso->get_display();
}


void dump_reloc_data_directory(byte * start, int size, image_section_helper & sh, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_reloc_directory * dso = new dbobj_reloc_directory(start, size, &cd, sh);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD5", dbp));

	std::cout << dso->get_display();
}



void dump_delayedimport_data_directory(byte * start, int size, image_section_helper & sh, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_delay_import_directory * dso = new dbobj_delay_import_directory(start, size, &cd, sh);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD13", dbp));

	std::cout << dso->get_display();
}

void dump_tls_data_directory(byte * start, int size, image_section_helper & sh, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_tls_directory * dso = new dbobj_tls_directory(start, size, &cd, sh);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD9", dbp));

	std::cout << dso->get_display();
}

void dump_bound_import_data_directory(byte * start, int size, result_heirachy_impl * prh)
{
	dbobj_bound_import_directory * dso = new dbobj_bound_import_directory(start, size);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dso->get_display();
}

void dump_debug_data_directory(byte * start, int size, image_section_helper & sh, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_debug_directory * dso = new dbobj_debug_directory(start, size, &cd);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD6", dbp));

	std::cout << dso->get_display();
}


void dump_com_descriptor_directory(byte * start, int size, image_section_helper & sh, cofpcdat & cd, result_heirachy_impl * prh)
{
	dbobj_com_descriptor_directory * dso = new dbobj_com_descriptor_directory(start, size, &cd);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node("DD14", dbp));

	std::cout << dso->get_display();
}

void dump_data_directory(int ddid, IMAGE_DATA_DIRECTORY * pd, cofpcdat & cd, result_heirachy_impl * prh)
{
	if (pd->Size == 0) return;

	if (ddid == IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT) 
	{
		dump_bound_import_data_directory(cd.base + pd->VirtualAddress, pd->Size, prh);
		return;
	}
	byte * dirrva = NULL;
	try 
	{
		dirrva = cd.get_rva(pd->VirtualAddress);
	}
	catch (invalid_rva_error & )
	{
		return;
	}
	image_section_helper sh = cd.get_rva_section(pd->VirtualAddress);

	switch (ddid)
	{
	case IMAGE_DIRECTORY_ENTRY_EXPORT:
		dump_export_data_directory(dirrva, pd->Size, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_IMPORT:
		dump_import_data_directory(dirrva, pd->Size, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_RESOURCE:
		// Not dumped by dumpbin
		break;
	case IMAGE_DIRECTORY_ENTRY_EXCEPTION:
		dump_exception_data_directory(dirrva, pd->Size, sh, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_SECURITY:
//		throw std::runtime_error("**Unhandled data directory IMAGE_DIRECTORY_ENTRY_SECURITY**\n");
		break;
	case IMAGE_DIRECTORY_ENTRY_BASERELOC:
		dump_reloc_data_directory(dirrva, pd->Size, sh, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_DEBUG:
		dump_debug_data_directory(dirrva, pd->Size, sh, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_ARCHITECTURE:
//      IMAGE_DIRECTORY_ENTRY_COPYRIGHT       7   // (X86 usage)
//		throw std::runtime_error("**Unhandled data directory IMAGE_DIRECTORY_ENTRY_ARCHITECTURE**\n");
		break;
	case IMAGE_DIRECTORY_ENTRY_GLOBALPTR:
		throw std::runtime_error("**Unhandled data directory IMAGE_DIRECTORY_ENTRY_GLOBALPTR**\n");
		break;
	case IMAGE_DIRECTORY_ENTRY_TLS:
		dump_tls_data_directory(dirrva, pd->Size, sh, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG:
		dump_load_config_data_directory(dirrva, pd->Size, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT:
		throw std::runtime_error("**CANT GET HERE **\n");
		break;
	case IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT:
		dump_delayedimport_data_directory(dirrva, pd->Size, sh, cd, prh);
		break;
	case IMAGE_DIRECTORY_ENTRY_IAT:
		// Dumpbin seems to skip!
		break;
	case IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR:
		dump_com_descriptor_directory(dirrva, pd->Size, sh, cd, prh);
		break;
	default:
		throw std::runtime_error("**Unhandled data directory**\n");
		break;
	}

}

void dump_linenumbers(IMAGE_SECTION_HEADER & h, int id, cofpcdat & cd, dbobj_coff_section * dco, result_heirachy_impl * prh)
{
	dbobj_section_linenumbers * dso = new dbobj_section_linenumbers(dco);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dso->get_display();
}


void dump_relocations(IMAGE_SECTION_HEADER & h, int id, cofpcdat & cd, dbobj_coff_section * dco, result_heirachy_impl * prh)
{
	dbobj_section_relocations * dso = new dbobj_section_relocations(dco);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dso->get_display();
}

void dump_section_header(IMAGE_SECTION_HEADER & h, int id, cofpcdat & cd, image_section_helper & sh, dbobj_coff_section * dco, result_heirachy_impl * prh)
{
	dbobj_section_header * dso = new dbobj_section_header(dco);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dso->get_display();
}

void dump_section_linker_directives(IMAGE_SECTION_HEADER & h, cofpcdat & cd, dbobj_coff_section * dco, result_heirachy_impl * prh)
{
	dbobj_section_linker_directives * dso = new dbobj_section_linker_directives(dco);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dso->get_display();
}

void dump_section_raw_data(IMAGE_SECTION_HEADER & h, int id, cofpcdat & cd, image_section_helper & sh, dbobj_coff_section * dco, result_heirachy_impl * prh)
{
	dbobj_section_data * dso = new dbobj_section_data(dco);
	dbobj::ptr dbp(dso);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dso->get_display();
}

void dump_section(IMAGE_SECTION_HEADER & h, int id, cofpcdat & cd, result_heirachy_impl * prh)
{
	image_section_helper sh(&h, id, cd.base);

	dbobj_coff_section * dco = new dbobj_coff_section(&h, id, &cd, sh);
	dbobj::ptr dbp(dco);
	rhi_stack rs(prh, prh->add_node(dbp));

	image_section_helper & firstsh = cd.get_sectionsbyname().find(sh.name())->second.front();

	unsigned long ignorediffmask = (IMAGE_SCN_ALIGN_MASK | IMAGE_SCN_LNK_COMDAT);
	if ((firstsh.characteristics() ^ sh.characteristics()) & ~ignorediffmask) 
	{
		std::stringstream ss;
		ss << "DUMPBIN : warning LNK4078: multiple '" << sh.name() << "' sections found with different attributes (" << lpad(hex(sh.characteristics()), 8, '0') << ")\n\n";
		// FIXME - DISPLAY
		dco->add_warning(ss.str());
		std::cout << ss.str();
	}

	dump_section_header(h, id, cd, sh, dco, prh);

	std::cout << "\n";
	if (h.SizeOfRawData && h.PointerToRawData)
	{
		dump_section_raw_data(h, id, cd, sh, dco, prh);
	}

	if ((h.Characteristics & IMAGE_SCN_LNK_INFO) && str(h.Name,IMAGE_SIZEOF_SHORT_NAME) == ".drectve")
	{
		dump_section_linker_directives(h, cd, dco, prh);
	}

	if (cd.has_data_directories(id))
	{
		rhi_stack rsc(prh, prh->add_node(dbobj::ptr(new dbobj_container("Data Directories"))));
		cofpcdat::datadirectoryvec & ddv = cd.get_data_directories(id);
		for (cofpcdat::datadirectoryvec::iterator vit = ddv.begin(); vit != ddv.end(); ++vit)
		{
			dump_data_directory(vit->ddid, vit->pdd, cd, prh);
		}
	}

	if (h.NumberOfRelocations)
	{
		dump_relocations(h, id, cd, dco, prh);
	}

	if (h.NumberOfLinenumbers)
	{
		dump_linenumbers(h, id, cd, dco, prh);
	}

}


std::map<int, int> extract_comdat_refs(IMAGE_SYMBOL * psyms, int nsyms)
{
	std::map<int, int> comdat_sym_map;
	std::map<int, int> count_seen;

	int idx;
	for (idx=0;idx<nsyms;++idx)
	{
		int thisidx = idx;
		IMAGE_SYMBOL & sym = psyms[idx];
		if (sym.NumberOfAuxSymbols) idx += sym.NumberOfAuxSymbols;
		if (sym.SectionNumber > 0)
		{
			int sn = sym.SectionNumber;
			/*
			 *
			 */
#if 0
			if (sym.Value == 0 && IMAGE_SYM_CLASS_EXTERNAL == sym.StorageClass) 
			{
				comdat_sym_map[sn] = thisidx;
			}
#else
			switch (count_seen[sn])
			{
			case 0:
				{
					// boost doesn't obey this!!!
					if (sym.Value == 0 && IMAGE_SYM_TYPE_NULL == sym.Type && 
						IMAGE_SYM_CLASS_STATIC == sym.StorageClass && sym.NumberOfAuxSymbols)
					{
						count_seen[sn]++;
					}
					else
					{
						continue;
					}
				}
				break;
			case 1:
				if (IMAGE_SYM_CLASS_LABEL != sym.StorageClass) count_seen[sn]++;
				comdat_sym_map[sn] = thisidx;
				break;
			default:
				break;
			}
#endif
		}
	}
	return comdat_sym_map;
}

void dump_symbol_table(cofpcdat & cd, result_heirachy_impl * prh)
{
	int nsym = cd.nsym;
	IMAGE_SYMBOL * psym = cd.psym;
	if (!nsym || !psym) return;

	dbobj_coff_symbol_table * dco = new dbobj_coff_symbol_table(cd);
	dbobj::ptr dbp(dco);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dco->get_display();
}


void dump_optional_header(IMAGE_OPTIONAL_HEADER32 * hdr, cofpcdat & cd, result_heirachy_impl * prh)
{
	cd.imagebase = hdr->ImageBase; 
	cd.alignment = hdr->SectionAlignment;

	dbobj_coff_opt_header * dco = new dbobj_coff_opt_header(hdr, &cd);
	dbobj::ptr dbp(dco);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dco->get_display();
	
	IMAGE_DATA_DIRECTORY * img = (IMAGE_DATA_DIRECTORY *)((byte*)&hdr->NumberOfRvaAndSizes + 4);

	if (hdr->NumberOfRvaAndSizes)
	{
		rhi_stack rsdd(prh, prh->add_node(dbobj::ptr(new dbobj_container("Data Directories"))));

		for (unsigned long i=0;i<hdr->NumberOfRvaAndSizes;++i,++img)
		{
			std::cout << dump_ddinfo(img, data_directory_name(i).c_str());
			if (img->Size) 
			{
				try {
					cd.add_data_directory(i, img);
					{
						std::stringstream ss;
						ss << "DD" << i;
						prh->add_node(dbobj::ptr(new dbobj_indirection(ss.str())));
					}
				} catch (invalid_rva_error & )
				{
					// We can handle invalid RVAs.
				}
			}
		}
	}

	std::cout << "\n";
}

void dump_optional_header(byte * data, cofpcdat & cd, result_heirachy_impl * prh)
{
	std::cout << "OPTIONAL HEADER VALUES\n";
	unsigned short magic = *(unsigned short*)data;
	switch (magic)
	{
	case IMAGE_NT_OPTIONAL_HDR32_MAGIC:
		dump_optional_header((IMAGE_OPTIONAL_HEADER32*)data, cd, prh);
		break;
	case IMAGE_NT_OPTIONAL_HDR64_MAGIC:
		std::cout << "optional HDR64 not handled\n";
		break;
	case IMAGE_ROM_OPTIONAL_HDR_MAGIC:
		std::cout << "optional ROM HDR not handled\n";
		break;
	default:
		std::cout << "Unknown optional header magic [" << magic << "]\n";
		break;
	}
	std::cout << "\n";
}

/*
 * This is now only pre processing.
 */
void dump_coff_summary(std::vector<cofpcdat> * pcov, result_heirachy_impl * prh)
{
	unsigned long ignorediffmask = (IMAGE_SCN_ALIGN_MASK | IMAGE_SCN_LNK_COMDAT);

	std::vector<cofpcdat> & cov = *pcov;

	std::map<std::pair<std::string, unsigned long>, unsigned long> aggres;
	std::map<std::pair<std::string, unsigned long>, unsigned long>::iterator ait;

	for (std::vector<cofpcdat>::iterator cvi = cov.begin(); cvi != cov.end(); ++cvi)
	{
		cofpcdat & cd = *cvi;
		cofpcdat::sectionsnamemap & nm = cd.get_sectionsbyname();

		cofpcdat::sectionsnamemap::iterator mit;
		for (mit = nm.begin(); mit != nm.end(); ++mit)
		{
			std::vector<image_section_helper> & shv = mit->second;
			for (std::vector<image_section_helper>::iterator mit2 = shv.begin(); mit2 != shv.end(); ++mit2)
			{
				image_section_helper & sh = *mit2;
				std::string name = mit->first;
				if (name[0] == '/') name = cd.stringtable_string(name);

				aggres[std::pair<std::string, unsigned long>(name, sh.characteristics() & ~ignorediffmask)] += sh.summarysize( cd.alignment );
			}
		}
	}

	dbobj_coff_summary * dsi = new dbobj_coff_summary();
	dbobj::ptr dbp(dsi);
	prh->add_node(dbp);

	for (ait = aggres.begin(); ait != aggres.end(); ++ait)
	{
		dsi->add_item(ait->second, ait->first.first);
	}

	std::cout << dsi->get_display();

	return;
}


void dump_coff_obj_imp_obj(byte * data, byte * base, std::vector<coff_export_symbol> & cev, result_heirachy_impl * prh, int identify_type)
{
	dbobj_coff_import_obj * dco = new dbobj_coff_import_obj(data);
	dbobj::ptr dbp(dco);
	rhi_stack rs(prh, prh->add_node(dbp));

	data += sizeof(IMPORT_OBJECT_HEADER);
	std::string name_imp = (char*)data;

	std::string undec = unmangle(name_imp);
	if (undec != "")
	{
		std::stringstream ss;
		ss << " (" << undec << ")";
		name_imp += ss.str();
	}
	coff_export_symbol ce;
	ce.symbol = name_imp;
	cev.push_back(ce);
}

void dump_coff_obj_file(byte * data, byte * base, unsigned long tot_size, std::vector<cofpcdat> & cov, result_heirachy_impl * prh, int ident_type)
{
	byte * data2 = data;
	IMAGE_FILE_HEADER * hdr = (IMAGE_FILE_HEADER *)data2;

	data2 += IMAGE_SIZEOF_FILE_HEADER;
	std::cout << image_header_file_type(hdr);

	dbobj_coff_file * dco = new dbobj_coff_file(hdr);
	dbobj::ptr dbp(dco);
	rhi_stack rs(prh, prh->add_node(dbp));

	dump_coff_header(*hdr, prh);

	// Dump optional header if it exists...
	cofpcdat ccd;
	cov.push_back(ccd);
	cofpcdat & cd  = cov.back();
	cd.base		   = base;
	cd.comdatrefs  = extract_comdat_refs((IMAGE_SYMBOL *)(data + hdr->PointerToSymbolTable), hdr->NumberOfSymbols);
	// Evidence of offset from base due to AppleMobileBackup.exe etc
	cd.psym		   = hdr->PointerToSymbolTable ? (IMAGE_SYMBOL *)(((hdr->Characteristics & IMAGE_FILE_EXECUTABLE_IMAGE) ? base : data) + hdr->PointerToSymbolTable) : 0;
	cd.nsym		   = hdr->NumberOfSymbols;
	cd.stringtable = (byte *)(cd.psym + cd.nsym);
	cd.preprocess_symtable();
	cd.ident_type  = ident_type;

	IMAGE_SECTION_HEADER * sections = (IMAGE_SECTION_HEADER*)(data2 + hdr->SizeOfOptionalHeader);
	// 
	// Pre-cook the virtual address ranges.
	for (int i=0;i<hdr->NumberOfSections;++i)
	{
		cook_section_information(sections[i], i+1, cd);
	}

	if (hdr->SizeOfOptionalHeader)
	{
		dump_optional_header(data2, cd, prh);
	}
	data2 += hdr->SizeOfOptionalHeader; // unneeded

	if (hdr->NumberOfSections)
	{
		rhi_stack rsc(prh, prh->add_node(dbobj::ptr(new dbobj_container("Sections"))));
		for (int i=0;i<hdr->NumberOfSections;++i)
		{
			dump_section(sections[i], i+1, cd, prh);
		}
	}
	// Dump symbol table.
	dump_symbol_table(cd, prh);	

	unsigned long strtsize = tot_size - (cd.stringtable - base);
	if (tot_size && strtsize && cd.psym)
	{
		std::cout << "String Table Size = 0x" << hex(*(long *)(cd.stringtable)) << " bytes\n\n";
	}

}

void dump_coff_obj(byte * base, int data_size, std::vector<coff_export_symbol> & cev, std::vector<cofpcdat> & cov, result_heirachy_impl * prh, int ident_type)
{
	IMAGE_FILE_HEADER * if_hdr = (IMAGE_FILE_HEADER *)base;
	if (if_hdr->Machine == IMAGE_FILE_MACHINE_UNKNOWN &&
	    if_hdr->NumberOfSections == IMPORT_OBJECT_HDR_SIG2)
	{
		dump_coff_obj_imp_obj(base, base, cev, prh, ident_type);
	}
	else
	{
		dump_coff_obj_file(base, base, data_size, cov, prh, ident_type);
	}
}

void dump_obj(byte * base, int len, result_heirachy_impl * prh)
{
	std::cout << "File Type: COFF OBJECT\n\n";

	std::vector<coff_export_symbol> cev;
	std::vector<cofpcdat>			cov;

	dbobj_object * ppe = new dbobj_object(base);
	dbobj::ptr sppe(ppe);
	rhi_stack rs(prh, prh->add_node(sppe));

	dump_coff_obj(base, len, ppe->cev(), ppe->cov(), prh, IDENTIFY_OBJ);
	dump_coff_summary(&cov, prh);
}

void dump_anon_obj(byte * base, int len, result_heirachy_impl * prh)
{
	dbobj_anonymous_object * dsi = new dbobj_anonymous_object((ANON_OBJECT_HEADER*)base);
	dbobj::ptr dbp(dsi);
	rhi_stack rs(prh, prh->add_node(dbp));

	std::cout << dsi->get_display();
}

void dump_library(byte * base, int maxbyt, result_heirachy_impl * prh)
{
	std::string scurrentlongname;
	std::cout << "File Type: LIBRARY\n\n";

	byte * data = base + IMAGE_ARCHIVE_START_SIZE;

	dbobj_library * dsi = new dbobj_library(data);
	dbobj::ptr dbp(dsi);
	rhi_stack rs(prh, prh->add_node(dbp));
	
	std::vector<coff_export_symbol> & cev = dsi->cev();
	std::vector<cofpcdat>			& cov = dsi->cov();

	// A little gross, but for simplicity we keep pointers into cov elsewhere.
	// This requires that we presize it, otherwise we'll end up with it moving underneath
	// us.... 

	int presize = 0;
	do 
	{
		++presize;

		if (data - base >= maxbyt) break;
		IMAGE_ARCHIVE_MEMBER_HEADER * hdr = (IMAGE_ARCHIVE_MEMBER_HEADER*)data;

		data += IMAGE_SIZEOF_ARCHIVE_MEMBER_HDR;

		// some padding here?
		int data_size = atoi(str(hdr->Size,10).c_str());
		data += data_size;
		unsigned long lda = (unsigned long)data;
		lda += 1;
		lda &= ~1;
		data = (byte *)lda;
	} while (true);
	cov.reserve(presize);

	// rewind.
	data = base + IMAGE_ARCHIVE_START_SIZE;
	int idx = 0;
	do 
	{
		++idx;

		if (data - base >= maxbyt) break;
		IMAGE_ARCHIVE_MEMBER_HEADER * hdr = (IMAGE_ARCHIVE_MEMBER_HEADER*)data;

		dbobj_image_archive_member_header * diam = new dbobj_image_archive_member_header(hdr, scurrentlongname, data-base, idx);
		dbobj::ptr dbp(diam);
		rhi_stack rs(prh, prh->add_node(dbp));

		std::cout << diam->get_display();

//		dump_image_archive_member_header(*hdr, scurrentlongname, data-base);
		scurrentlongname = "";
		data += IMAGE_SIZEOF_ARCHIVE_MEMBER_HDR;

		byte * data2 = data;
		// some padding here?
		int data_size = atoi(str(hdr->Size,10).c_str());
		data += data_size;
		unsigned long lda = (unsigned long)data;
		lda += 1;
		lda &= ~1;
		data = (byte *)lda;
		// Linker member.


		if (!memcmp(hdr->Name, IMAGE_ARCHIVE_LINKER_MEMBER, 16))
		{
			switch (idx)
			{
			case 1:
				{
					dbobj_iamh_symbols * diam = new dbobj_iamh_symbols(data2);
					dbobj::ptr dbp(diam);
					prh->add_node(dbp);

					std::cout << diam->get_display();
				}
				break;
			case 2:
				{
					dbobj_iamh_offsets * diam = new dbobj_iamh_offsets(data2);
					dbobj::ptr dbp(diam);
					prh->add_node(dbp);

					std::cout << diam->get_display();
				}
				break;
			}
		
		}
		else if (!memcmp(hdr->Name, IMAGE_ARCHIVE_LONGNAMES_MEMBER, 16))
		{
			scurrentlongname = std::string((char*)data2);
		}
		else
		{
			dump_coff_obj(data2, data_size, cev, cov, prh, IDENTIFY_LIB);
		}
	} while (true);

	if (!cev.empty())
	{
		dbobj_coff_exports * dexp = new dbobj_coff_exports(cev);
		dbobj::ptr dbp(dexp);
		prh->add_node(dbp);

		std::cout << dexp->get_display();
	}

	if (!cov.empty())
	{
		dump_coff_summary(&cov, prh);
	}
}

byte * pestart(byte * data, int len)
{
	if (len <= 0x3c) return 0;
	unsigned long offs = (*(long*)(data + 0x3c));
	if (offs > (unsigned long)len) return false;
	if (memcmp(data + offs, "PE\0\0", 4)) return 0;
	return data+offs;
}

void dump_pe(byte * base, unsigned long totsize, result_heirachy_impl * prh)
{
	byte * data = pestart(base, totsize);
	std::cout << "PE signature found\n\n";
	data += 4;

	// Keep hold of native dbobj_pe so we can get at COV.
	dbobj_pe * ppe = new dbobj_pe(data);
	dbobj::ptr sppe(ppe);
	rhi_stack rs(prh, prh->add_node(sppe));

	std::vector<cofpcdat> & cov = ppe->cov();
	
	dump_coff_obj_file(data, base, totsize, cov, prh, IDENTIFY_PE);

	if (!cov.empty())
	{
		dump_coff_summary(&cov, prh);
	}
}

/*
 * Obviously, this isn't the MS application!  However for the purposes of diffing the output of the two, 
 * this is a copy of the header.....
 */
void dump_tmpheader(std::string fname)
{
	std::cout << "Microsoft (R) COFF/PE Dumper Version 8.00.50727.42\n";
	std::cout << "Copyright (C) Microsoft Corporation.  All rights reserved.\n";
	std::cout << "\n\n";
	std::cout << "Dump of file " << fname << "\n";
}



int identify(byte * base, int len)
{
	if (!memcmp(base, IMAGE_ARCHIVE_START, IMAGE_ARCHIVE_START_SIZE))
	{
		return IDENTIFY_LIB;
	}
	else if (0 != (pestart(base, len)))
	{
		return IDENTIFY_PE;
	}

	// trickier! :(
	IMAGE_FILE_HEADER * ifh = (IMAGE_FILE_HEADER*)base;
	if (machine_type(ifh->Machine) != "???")
	{
		return IDENTIFY_OBJ;
	}

	ANON_OBJECT_HEADER * aoh = (ANON_OBJECT_HEADER*)base;
	if (aoh->Sig1 == IMAGE_FILE_MACHINE_UNKNOWN && 
		aoh->Sig2 == 0xFFFF)
	{
		return IDENTIFY_ANON_OBJ;
	}

	if (*(short*)base == 0x5a4d) return IDENTIFY_OLDDOS;
//	hexdump(base, len < 1000 ? len : 1000);

	return IDENTIFY_UNKNOWN;
}

_declspec(dllexport) result_heirachy::ptr process_file(std::string fname, bool dumptoconsole)
{
	result_heirachy_impl * rhi = new result_heirachy_impl();
	result_heirachy::ptr rh(rhi);

	rhi->set_filedata(memory_mapped_file::ptr(new memory_mapped_file(fname.c_str())));
	memory_mapped_file & mmf = *(rhi->get_filedata());
	if (!mmf.is_valid()) 
	{
		std::cerr << "Failed.\n";
		return result_heirachy::ptr(new result_heirachy_error("Failed to load file"));
	}
	
	byte * base = reinterpret_cast<byte *>(mmf.get_base());

	fname = tolower(fname);

	try {

		dump_tmpheader(fname);
		std::cout << "\n";

		switch (identify(base, mmf.get_size()))
		{
		case IDENTIFY_LIB:
			dump_library(base, mmf.get_size(), rhi);
			break;
		case IDENTIFY_PE:
			dump_pe(base, mmf.get_size(), rhi);
			break;
		case IDENTIFY_OBJ:
			dump_obj(base, mmf.get_size(), rhi);
			break;
		case IDENTIFY_ANON_OBJ:
			dump_anon_obj(base, mmf.get_size(), rhi);
			break;
		case IDENTIFY_OLDDOS:
			{
				std::stringstream ssc;
				ssc << "DUMPBIN : warning LNK4094: '" << fname << "' is an MS-DOS executable; use EXEHDR to dump it\n\n";	
				ssc << "  Summary\n\n";
				return result_heirachy::ptr(new result_heirachy_error(ssc.str()));
			}
			break;
		default:
			{
				std::stringstream ssc;
				ssc << fname << " : warning LNK4048: Invalid format file; ignored\n\n";
				ssc << "  Summary\n\n";
				return result_heirachy::ptr(new result_heirachy_error(ssc.str()));
			}
		}
	}
	catch (std::runtime_error & e)
	{
		std::stringstream ssc;
		ssc << "** FATAL ERROR : " << e.what() << " **\n";
		return result_heirachy::ptr(new result_heirachy_error(ssc.str()));
	}
	catch (...)
	{
		std::stringstream ssc;
		ssc << "Unknown error.";
		return result_heirachy::ptr(new result_heirachy_error(ssc.str()));
	}
	return rh;
}

/* 
 * If we were going to wrap this in an incredibly simple console app..... it'd look like this ;)
 */
#ifdef _DEBUG
void usage()
{
	std::cout << "Usage: \n";
}

int _tmain(int argc, _TCHAR* argv[])
{
	if (argc < 2) 
	{
		usage();
		return 0;
	}
	process_file(std::string(argv[1]), true);
	return 1;
}
#endif